use hashbrown::HashMap;
use lws_client::lws_vfs_client::LwsVfsClient;
use lws_client::{
    Access, Create, FileInfo, Flush, Fstat, GetConfig, Getattr, HelloRequest, Mkdir, Open, Opendir,
    Read, Readdir, Release, Releasedir, Rename, Rmdir, Timespec, Truncate, Unlink, Utimens, Write,
};
use serde_json::{self, Value};
use std::error::Error;
use std::ffi::OsStr;
use std::io::{Error as IoError, ErrorKind, Read as IoRead};
use std::thread::{self};
use std::{env, vec};
use tokio::runtime::{Builder, Runtime};
use tonic::transport::Channel as RpcChannel;
extern crate log;

pub mod lws_client {
    tonic::include_proto!("lws_vfs"); //导入lws vfs proto buffer
}

#[derive(Debug, Default)]
pub struct Config {
    mount_map: HashMap<String, String>,
}
impl Config {
    pub fn new(json: &str) -> Result<Config, Box<dyn Error>> {
        let json: Value = serde_json::from_str(json)?;
        let mounts = match json.get("mount_map") {
            Some(mounts) => mounts.as_object().unwrap(),
            None => {
                return Err(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "no mount map",
                )))
            }
        };
        let mount_map = mounts
            .iter()
            .map(|(key, value)| {
                (
                    key.as_str().to_string(),
                    value.as_str().unwrap().to_string(),
                )
            })
            .collect();
        Ok(Config { mount_map })
    }

    pub fn get_mount(&self) -> Vec<&String> {
        let mut ret = vec![];
        for (_path, mount) in &self.mount_map {
            ret.push(mount);
        }
        ret
    }
}
use std::sync::{Arc, Condvar, Mutex, RwLock};
struct RpcPool {
    clients: Arc<Mutex<Vec<Arc<RwLock<Rpc>>>>>,
    index: Arc<RwLock<usize>>,
    wait: Arc<(Mutex<()>, Condvar)>,
}

struct Rpc {
    pub client: LwsVfsClient<RpcChannel>,
    pub rt: Arc<Runtime>,
    pub in_use: bool,
    pub tx: Arc<mpsc::Sender<FileAttrCacheMsg>>,
    pub rx: Arc<Mutex<mpsc::Receiver<FileAttrCacheMsg>>>,
}

impl RpcPool {
    async fn new(size: usize, uri: String) -> Result<Self, Box<dyn std::error::Error>> {
        let mut clients = Vec::with_capacity(size);
        let uri = uri.to_ascii_lowercase();
        log::info!("create rpc pool size: {}", size);
        let channel = RpcChannel::from_shared(uri)?.connect().await?;
        for _ in 0..size {
            let client = LwsVfsClient::new(channel.clone());
            let (tx, rx) = mpsc::channel();
            clients.push(Arc::new(RwLock::new(Rpc {
                client,
                in_use: false,
                rt: Arc::new(
                    Builder::new_multi_thread()
                        .worker_threads(1)
                        .build()
                        .unwrap(),
                ),
                tx: Arc::new(tx),
                rx: Arc::new(Mutex::new(rx)),
            })));
        }
        let wait = (Mutex::new(()), Condvar::new());
        Ok(Self {
            clients: Arc::new(Mutex::new(clients)),
            index: Arc::new(RwLock::new(0)),
            wait: Arc::new(wait),
        })
    }

    fn get(&self) -> Result<Arc<RwLock<Rpc>>, Box<dyn Error>> {
        // TODO: 优化遍历的逻辑
        let find_idle = |rpc_pool: &Self| -> Result<Arc<RwLock<Rpc>>, Box<dyn Error>> {
            let cs = self.clients.clone();
            let clients = &(cs.lock().unwrap());
            let size = clients.len();
            let index = *rpc_pool.index.read().unwrap();
            for i in 0..size {
                let idx = (index + i) % size;
                let entry = clients[idx].read().unwrap();
                if !entry.in_use {
                    drop(entry);
                    let mut entry: std::sync::RwLockWriteGuard<'_, Rpc> =
                        clients[idx].write().unwrap();
                    entry.in_use = true;
                    return Ok(clients[idx].clone());
                }
            }
            Err(Box::new(IoError::new(
                ErrorKind::Other,
                "get free rpc client fail",
            )))
        };
        loop {
            if let Ok(rpc) = find_idle(self) {
                return Ok(rpc);
            }
            let (mutex, cond) = &(*self.wait.clone());
            let timeout = Duration::from_millis(100);
            let (_unused, timeout_res) = cond.wait_timeout(mutex.lock().unwrap(), timeout).unwrap();
            if timeout_res.timed_out() {
                break;
            }
        }
        log::warn!("get free rpc client time out");
        Err(Box::new(IoError::new(
            ErrorKind::Other,
            "get free rpc client time out",
        )))
    }

    fn put(&self, client: Arc<RwLock<Rpc>>) {
        let mut i = 0;
        for entry in self.clients.lock().unwrap().iter_mut() {
            if Arc::ptr_eq(&entry, &client) {
                let mut item = entry.write().unwrap();
                item.in_use = false;
                *self.index.write().unwrap() = i;
                let (mutex, cond) = &(*self.wait.clone());
                let _unused = mutex.lock().unwrap();
                cond.notify_one();
                break;
            }
            i += 1;
        }
    }
}

struct VirFs {
    _name: String,
    stat: lws_client::Fstat,
    sub: HashMap<String, VirFs>,
    is_root: bool,
}

impl VirFs {
    fn create<T>(name: T, is_dir: bool) -> VirFs
    where
        T: ToString,
    {
        let now = SystemTime::now()
            .duration_since(SystemTime::UNIX_EPOCH)
            .unwrap()
            .as_nanos() as u64;
        VirFs {
            _name: name.to_string(),
            stat: lws_client::Fstat {
                fst_mode: 0o0555 | {
                    if is_dir {
                        libc::S_IFDIR
                    } else {
                        0
                    }
                } as u64,
                fst_ino: 0,
                fst_dev: 0,
                fst_rdev: 0,
                fst_nlink: 0,
                fst_uid: 0,
                fst_gid: 0,
                fst_size: 0,
                fst_atime: now,
                fst_mtime: now,
                fst_ctime: now,
                fst_blksize: 0,
                fst_blocks: 0,
            },
            sub: HashMap::new(),
            is_root: true,
        }
    }
    pub fn new() -> VirFs {
        let mut node = VirFs::create("/", true);
        node.is_root = true;
        node
    }

    pub fn add<T>(&mut self, name: T)
    where
        T: ToString,
    {
        let name = name.to_string();
        let ret = name
            .split('/')
            .filter(|x| x.len() > 0)
            .map(|y| y)
            .collect::<Vec<&str>>();
        let name = ret[0];
        log::debug!("add fs {}", name);
        let mut node = VirFs::create(name.to_string(), true);
        node.is_root = false;
        self.sub.insert(name.to_string(), node);
    }

    pub fn find<T>(&self, name: T) -> Option<&VirFs>
    where
        T: ToString,
    {
        let name = name.to_string();
        let mut vname = name.split('/').filter(|x| x.len() > 0);
        //log::info!("try match {}", name);
        if let Some(_curr) = vname.next() {
            return None;
        }
        Some(self)
        //self.match_path(&mut vname)
    }

    pub fn match_path<'a>(&self, name_list: &mut impl Iterator<Item = &'a str>) -> Option<&VirFs> {
        if let Some(current) = name_list.next() {
            match self.sub.get(current) {
                Some(v) => {
                    // try match sub dir
                    if let Some(sub) = v.match_path(name_list) {
                        Some(sub)
                    // sub dirs not match
                    } else {
                        None
                    }
                }
                None => None,
            }
        } else {
            // find over
            Some(self)
        }
    }

    pub fn readdir(&self) -> Vec<String> {
        let mut ret = vec![];
        for name in self.sub.keys() {
            ret.push(name.to_string());
        }
        log::trace!("readdir: {:?}", ret);
        ret
    }
}

type FileAttrInfo = Result<FileAttr, i32>;
pub struct FileAttrCache {
    attr: FileAttrInfo,
    exp_time: u32,
}
use std::time::UNIX_EPOCH;
fn current_time_in_ms() -> u32 {
    let start = SystemTime::now();
    start
        .duration_since(UNIX_EPOCH)
        .expect("Time went backwards")
        .as_millis() as u32
}
impl FileAttrCache {
    pub fn new(attr: FileAttrInfo, cache_life: u32) -> FileAttrCache {
        FileAttrCache {
            attr,
            exp_time: current_time_in_ms() + cache_life,
        }
    }
}
use std::sync::mpsc;
pub enum FileAttrCacheMsg {
    Get(String, Arc<mpsc::Sender<FileAttrCacheMsg>>),
    GetReply(String, Option<FileAttrInfo>),
    Set(String, FileAttrInfo),
    Clr(Vec<String>),
}
pub struct FileAttrCacheCtx {
    pub cache: HashMap<String, FileAttrCache>,
    pub cache_size: u32,
    rx: Arc<Mutex<mpsc::Receiver<FileAttrCacheMsg>>>,
}

impl FileAttrCacheCtx {
    pub fn new(rx: mpsc::Receiver<FileAttrCacheMsg>) -> FileAttrCacheCtx {
        FileAttrCacheCtx {
            cache: HashMap::new(),
            cache_size: 0,
            rx: Arc::new(Mutex::new(rx)),
        }
    }
}

pub struct FileAttrCacheManager {
    handle: thread::JoinHandle<()>,
    tx: mpsc::Sender<FileAttrCacheMsg>,
}

impl FileAttrCacheManager {
    pub fn new(cache_life: u32) -> FileAttrCacheManager {
        let (tx, rx) = mpsc::channel();
        let cache_ctx = FileAttrCacheCtx::new(rx);
        let handle = thread::spawn(move || {
            Self::cache_manager(cache_ctx, cache_life);
        });
        FileAttrCacheManager { handle, tx }
    }

    fn cache_manager(mut ctx: FileAttrCacheCtx, cache_life: u32) {
        let rx = ctx.rx.lock().unwrap();
        fn invalid_cache(cache: &mut HashMap<String, FileAttrCache>) -> u32 {
            let mut cache_size = 0;
            let curr = current_time_in_ms();
            let del: Vec<String> = cache
                .iter()
                .filter(|&(_, val)| val.exp_time < curr)
                .map(|(key, _)| key.clone())
                .collect();
            for key in del {
                cache.remove(&key);
                log::debug!("invalid cache: {}", key);
                cache_size += 1;
            }
            cache_size
        }
        loop {
            match rx.recv_timeout(Duration::from_millis(200)) {
                Ok(req) => {
                    if ctx.cache_size > 256 {
                        ctx.cache_size -= invalid_cache(&mut ctx.cache);
                    }
                    match req {
                        FileAttrCacheMsg::Set(path, attr) => {
                            ctx.cache.insert(path, FileAttrCache::new(attr, cache_life));
                            ctx.cache_size += 1;
                        }
                        FileAttrCacheMsg::Get(path, chn) => match ctx.cache.get(&path) {
                            Some(attr) => {
                                let _ = chn.send(FileAttrCacheMsg::GetReply(path, Some(attr.attr)));
                            }
                            None => {
                                let _ = chn.send(FileAttrCacheMsg::GetReply(path, None));
                            }
                        },
                        FileAttrCacheMsg::Clr(paths) => {
                            for path in paths.into_iter() {
                                ctx.cache.remove(&path);
                                ctx.cache_size -= 1;
                            }
                        }
                        _ => {}
                    }
                }
                Err(_) => {
                    ctx.cache_size -= invalid_cache(&mut ctx.cache);
                }
            }
        }
    }

    pub fn set(&self, path: &String, attr: FileAttrInfo) {
        let _ = self.tx.send(FileAttrCacheMsg::Set(path.to_string(), attr));
    }

    pub fn get(
        &self,
        path: &String,
        tx: &Arc<mpsc::Sender<FileAttrCacheMsg>>,
        rx: &Arc<Mutex<mpsc::Receiver<FileAttrCacheMsg>>>,
    ) -> Option<FileAttrInfo> {
        //send get gile attribute request
        let _ = self
            .tx
            .send(FileAttrCacheMsg::Get(path.to_string(), tx.clone()));
        let total_timeout = Duration::from_millis(100);
        let start = Instant::now();
        loop {
            let elapsed = start.elapsed();
            if elapsed >= total_timeout {
                return None;
            }
            let timeout = total_timeout - elapsed;
            let resp = {
                match rx.lock().unwrap().recv_timeout(timeout) {
                    Ok(resp) => resp,
                    Err(_e) => return None,
                }
            };
            match resp {
                FileAttrCacheMsg::GetReply(rpath, resp) => {
                    if *path == rpath {
                        //message is not match
                        return resp;
                    }
                }
                _ => break,
            }
        }
        None
    }

    pub fn clr(&self, paths: Vec<String>) {
        let _ = self.tx.send(FileAttrCacheMsg::Clr(paths));
    }
}

pub struct LwsVfsIns {
    pub config: Config,
    rpcs: RpcPool,
    vir_root: VirFs,
    cache: FileAttrCacheManager,
}

macro_rules! ret2result {
    ($ret:expr) => {
        match $ret {
            0 => Ok(()),
            _ => Err($ret),
        }
    };
}
fn mode_to_filetype(mode: libc::mode_t) -> FileType {
    match mode & libc::S_IFMT {
        libc::S_IFDIR => FileType::Directory,
        libc::S_IFREG => FileType::RegularFile,
        libc::S_IFLNK => FileType::Symlink,
        libc::S_IFBLK => FileType::BlockDevice,
        libc::S_IFCHR => FileType::CharDevice,
        libc::S_IFIFO => FileType::NamedPipe,
        libc::S_IFSOCK => FileType::Socket,
        _ => {
            panic!("unknown file type {}", mode);
        }
    }
}

use fuse_mt::*;
use std::path::Path;
use std::time::{Duration, Instant, SystemTime};
impl FilesystemMT for LwsVfsIns {
    /// Called on mount, before any other function.
    fn init(&self, _req: RequestInfo) -> ResultEmpty {
        match self.init() {
            Ok(()) => {
                log::info!("init vir fs success");
                Ok(())
            }
            Err(e) => {
                log::error!("hello err: {:?}", e);
                Err(libc::ENOMSG)
            }
        }
    }

    /// Called on filesystem unmount.
    fn destroy(&self) {}

    /// Get the attributes of a filesystem entry.
    ///
    /// * `fh`: a file handle if this is called on an open file.
    fn getattr(&self, req: RequestInfo, path: &Path, fh: Option<u64>) -> ResultEntry {
        let time = |timens| {
            let secs = timens / (1000 * 1000);
            let nanos = timens % (1000 * 1000);
            SystemTime::UNIX_EPOCH + Duration::new(secs as u64, nanos as u32)
        };
        let file_attr = |fst: &Fstat, req: &RequestInfo| FileAttr {
            size: fst.fst_size,
            blocks: fst.fst_blocks,
            atime: time(fst.fst_atime),
            mtime: time(fst.fst_mtime),
            ctime: time(fst.fst_ctime),
            crtime: SystemTime::UNIX_EPOCH,
            kind: mode_to_filetype(fst.fst_mode as u32),
            perm: fst.fst_mode as u16,
            nlink: fst.fst_nlink as u32,
            uid: req.uid,
            gid: req.gid,
            rdev: fst.fst_rdev as u32,
            flags: 0,
        };

        let path = path.to_string_lossy().into_owned();
        if let Some(node) = self.vir_root.find(&path) {
            return Ok((Duration::new(0, 0), file_attr(&node.stat, &req)));
        }
        log::trace!("lws getattr {:?}", path);
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            if let Some(attr) = self.cache.get(&path, &rpc.tx, &rpc.rx) {
                let d = client.clone();
                drop(rpc);
                self.rpcs.put(d);
                match attr {
                    Ok(attr) => return Ok((Instant::now().elapsed(), attr)),
                    Err(_e) => {
                        return Err(libc::ENOENT);
                    }
                }
            }
            let request = tonic::Request::new(Getattr {
                path: path.to_string(),
                fi: Some(FileInfo {
                    fh: match fh {
                        Some(f) => f,
                        None => 0,
                    },
                    ..FileInfo::default()
                }),
                ..Getattr::default()
            });
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fgetattr(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            self.cache.set(&path, Err(resp.ret));
            return Err(resp.ret);
        }
        let attr = file_attr(&resp.stat.unwrap(), &req);
        self.cache.set(&path, Ok(attr.clone()));
        Ok((Instant::now().elapsed(), attr))
    }

    // The following operations in the FUSE C API are all one kernel call: setattr
    // We split them out to match the C API's behavior.

    /// Change the mode of a filesystem entry.
    ///
    /// * `fh`: a file handle if this is called on an open file.
    /// * `mode`: the mode to change the file to.
    fn chmod(&self, _req: RequestInfo, _path: &Path, _fh: Option<u64>, _mode: u32) -> ResultEmpty {
        Ok(())
    }

    /// Change the owner UID and/or group GID of a filesystem entry.
    ///
    /// * `fh`: a file handle if this is called on an open file.
    /// * `uid`: user ID to change the file's owner to. If `None`, leave the UID unchanged.
    /// * `gid`: group ID to change the file's group to. If `None`, leave the GID unchanged.
    fn chown(
        &self,
        _req: RequestInfo,
        _path: &Path,
        _fh: Option<u64>,
        _uid: Option<u32>,
        _gid: Option<u32>,
    ) -> ResultEmpty {
        Ok(())
    }

    /// Set the length of a file.
    ///
    /// * `fh`: a file handle if this is called on an open file.
    /// * `size`: size in bytes to set as the file's length.
    fn truncate(&self, _req: RequestInfo, path: &Path, _fh: Option<u64>, size: u64) -> ResultEmpty {
        let request = tonic::Request::new(Truncate {
            path: path.to_str().unwrap().into(),
            size: size as i64,
            ..Truncate::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.ftruncate(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Set timestamps of a filesystem entry.
    ///
    /// * `fh`: a file handle if this is called on an open file.
    /// * `atime`: the time of last access.
    /// * `mtime`: the time of last modification.
    fn utimens(
        &self,
        _req: RequestInfo,
        path: &Path,
        _fh: Option<u64>,
        atime: Option<SystemTime>,
        mtime: Option<SystemTime>,
    ) -> ResultEmpty {
        let time = move |time| -> Timespec {
            let time = match time {
                Some(time) => time,
                None => SystemTime::now(),
            };
            match time.duration_since(SystemTime::UNIX_EPOCH) {
                Ok(dur) => Timespec {
                    tv_sec: dur.as_secs() as i32,
                    tv_nsec: dur.subsec_nanos() as i64,
                },
                Err(_e) => Timespec::default(),
            }
        };
        let at = time(atime);
        let mt = time(mtime);

        let request = tonic::Request::new(Utimens {
            path: path.to_str().unwrap().into(),
            ts: vec![at, mt],
            ..Utimens::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.futimens(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    // END OF SETATTR FUNCTIONS

    /// Read a symbolic link.
    fn readlink(&self, _req: RequestInfo, _path: &Path) -> ResultData {
        Err(libc::ENOSYS)
    }

    /// Create a special file.
    ///
    /// * `parent`: path to the directory to make the entry under.
    /// * `name`: name of the entry.
    /// * `mode`: mode for the new entry.
    /// * `rdev`: if mode has the bits `S_IFCHR` or `S_IFBLK` set, this is the major and minor numbers for the device file. Otherwise it should be ignored.
    fn mknod(
        &self,
        _req: RequestInfo,
        _parent: &Path,
        _name: &OsStr,
        _mode: u32,
        _rdev: u32,
    ) -> ResultEntry {
        Err(libc::ENOSYS)
    }

    /// Create a directory.
    ///
    /// * `parent`: path to the directory to make the directory under.
    /// * `name`: name of the directory.
    /// * `mode`: permissions for the new directory.
    fn mkdir(&self, req: RequestInfo, parent: &Path, name: &OsStr, mode: u32) -> ResultEntry {
        let path = format!("{}/{}", parent.to_str().unwrap(), name.to_str().unwrap());
        if let Some(_unused) = self.vir_root.find(&path) {
            return Err(libc::ENOSYS);
        }
        let request = tonic::Request::new(Mkdir {
            path: path.to_string(),
            mode,
            ..Mkdir::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fmkdir(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            return Err(resp.ret)
        }
        self.getattr(req, Path::new(&path), None)
    }

    /// Remove a file.
    ///
    /// * `parent`: path to the directory containing the file to delete.
    /// * `name`: name of the file to delete.
    fn unlink(&self, _req: RequestInfo, parent: &Path, name: &OsStr) -> ResultEmpty {
        let path = format!("{}/{}", parent.to_str().unwrap(), name.to_str().unwrap());
        let request = tonic::Request::new(Unlink {
            path: path.to_string(),
            ..Unlink::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.funlink(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        self.cache.clr(vec![path]);
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Remove a directory.
    ///
    /// * `parent`: path to the directory containing the directory to delete.
    /// * `name`: name of the directory to delete.
    fn rmdir(&self, _req: RequestInfo, parent: &Path, name: &OsStr) -> ResultEmpty {
        let path = format!("{}/{}", parent.to_str().unwrap(), name.to_str().unwrap());
        if let Some(_unused) = self.vir_root.find(&path) {
            return Err(libc::ENOMSG);
        }
        let request = tonic::Request::new(Rmdir {
            path: path.to_string(),
            ..Rmdir::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.frmdir(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Create a symbolic link.
    ///
    /// * `parent`: path to the directory to make the link in.
    /// * `name`: name of the symbolic link.
    /// * `target`: path (may be relative or absolute) to the target of the link.
    fn symlink(
        &self,
        _req: RequestInfo,
        _parent: &Path,
        _name: &OsStr,
        _target: &Path,
    ) -> ResultEntry {
        Err(libc::ENOSYS)
    }

    /// Rename a filesystem entry.
    ///
    /// * `parent`: path to the directory containing the existing entry.
    /// * `name`: name of the existing entry.
    /// * `newparent`: path to the directory it should be renamed into (may be the same as `parent`).
    /// * `newname`: name of the new entry.
    fn rename(
        &self,
        _req: RequestInfo,
        parent: &Path,
        name: &OsStr,
        _newparent: &Path,
        _newname: &OsStr,
    ) -> ResultEmpty {
        let path = format!("{}/{}", parent.to_str().unwrap(), name.to_str().unwrap());
        if let Some(_unused) = self.vir_root.find(&path) {
            return Err(libc::ENOMSG);
        }
        let request = tonic::Request::new(Rename {
            path: path.to_string(),
            ..Rename::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.frename(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Create a hard link.
    ///
    /// * `path`: path to an existing file.
    /// * `newparent`: path to the directory for the new link.
    /// * `newname`: name for the new link.
    fn link(
        &self,
        _req: RequestInfo,
        _path: &Path,
        _newparent: &Path,
        _newname: &OsStr,
    ) -> ResultEntry {
        Err(libc::ENOSYS)
    }

    /// Open a file.
    ///
    /// * `path`: path to the file.
    /// * `flags`: one of `O_RDONLY`, `O_WRONLY`, or `O_RDWR`, plus maybe additional flags.
    ///
    /// Return a tuple of (file handle, flags). The file handle will be passed to any subsequent
    /// calls that operate on the file, and can be any value you choose, though it should allow
    /// your filesystem to identify the file opened even without any path info.
    fn open(&self, _req: RequestInfo, path: &Path, flags: u32) -> ResultOpen {
        let request = tonic::Request::new(Open {
            path: path.to_string_lossy().into_owned(),
            fi: Some(FileInfo {
                flags,
                ..FileInfo::default()
            }),
            ..Open::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fopen(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            return Err(resp.ret)
        }
        let fi = resp.fi.unwrap();
        Ok((fi.fh, fi.flags))
    }

    /// Read from a file.
    ///
    /// Note that it is not an error for this call to request to read past the end of the file, and
    /// you should only return data up to the end of the file (i.e. the number of bytes returned
    /// will be fewer than requested; possibly even zero). Do not extend the file in this case.
    ///
    /// * `path`: path to the file.
    /// * `fh`: file handle returned from the `open` call.
    /// * `offset`: offset into the file to stasync_rt reading.
    /// * `size`: number of bytes to read.
    /// * `callback`: a callback that must be invoked to return the result of the operation: either
    ///    the result data as a slice, or an error code.
    ///
    /// Return the return value from the `callback` function.
    fn read(
        &self,
        _req: RequestInfo,
        path: &Path,
        fh: u64,
        offset: u64,
        size: u32,
        callback: impl FnOnce(ResultSlice<'_>) -> CallbackResult,
    ) -> CallbackResult {
        let request = tonic::Request::new(Read {
            path: path.to_string_lossy().into_owned(),
            fi: Some(FileInfo {
                fh,
                ..FileInfo::default()
            }),
            size: size.into(),
            offset,
            ..Read::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return callback(Err(libc::ENOMSG)),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fread(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return callback(Err(libc::ENOMSG));
            }
        };
        log::trace!("get fread resp size is: {}", resp.size);
        if resp.ret != 0 {
            return callback(Err(resp.ret));
        }
        callback(Ok(&resp.buff))
    }

    /// Write to a file.
    ///
    /// * `path`: path to the file.
    /// * `fh`: file handle returned from the `open` call.
    /// * `offset`: offset into the file to stasync_rt writing.
    /// * `data`: the data to write
    /// * `flags`:
    ///
    /// Return the number of bytes written.
    fn write(
        &self,
        _req: RequestInfo,
        path: &Path,
        fh: u64,
        offset: u64,
        data: Vec<u8>,
        _flags: u32,
    ) -> ResultWrite {
        let request = tonic::Request::new(Write {
            path: path.to_string_lossy().into_owned(),
            fi: Some(FileInfo {
                fh,
                ..FileInfo::default()
            }),
            offset,
            size: data.len() as u64,
            buff: data,
            ..Write::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fwrite(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            return Err(resp.ret)
        }
        Ok(resp.size as u32)
    }

    /// Called each time a program calls `close` on an open file.
    ///
    /// Note that because file descriptors can be duplicated (by `dup`, `dup2`, `fork`) this may be
    /// called multiple times for a given file handle. The main use of this function is if the
    /// filesystem would like to return an error to the `close` call. Note that most programs
    /// ignore the return value of `close`, though.
    ///
    /// * `path`: path to the file.
    /// * `fh`: file handle returned from the `open` call.
    /// * `lock_owner`: if the filesystem supports locking (`setlk`, `getlk`), remove all locks
    ///   belonging to this lock owner.
    fn flush(&self, _req: RequestInfo, path: &Path, fh: u64, _lock_owner: u64) -> ResultEmpty {
        let request = tonic::Request::new(Flush {
            path: path.to_string_lossy().into_owned(),
            fi: Some(FileInfo {
                fh,
                ..FileInfo::default()
            }),
            ..Flush::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fflush(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Called when an open file is closed.
    ///
    /// There will be one of these for each `open` call. After `release`, no more calls will be
    /// made with the given file handle.
    ///
    /// * `path`: path to the file.
    /// * `fh`: file handle returned from the `open` call.
    /// * `flags`: the flags passed when the file was opened.
    /// * `lock_owner`: if the filesystem supports locking (`setlk`, `getlk`), remove all locks
    ///   belonging to this lock owner.
    /// * `flush`: whether pending data must be flushed or not.
    fn release(
        &self,
        _req: RequestInfo,
        path: &Path,
        fh: u64,
        _flags: u32,
        _lock_owner: u64,
        flush: bool,
    ) -> ResultEmpty {
        let path = path.to_string_lossy().into_owned();
        let request = tonic::Request::new(Release {
            path: path.to_string(),
            fi: Some(FileInfo {
                fh,
                ..FileInfo::default()
            }),
            flush: flush as u32,
            ..Release::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.frelease(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
        // self.cache.clr(vec![path]);
    }

    /// Write out any pending changes of a file.
    ///
    /// When this returns, data should be written to persistent storage.
    ///
    /// * `path`: path to the file.
    /// * `fh`: file handle returned from the `open` call.
    /// * `datasync`: if `false`, also write metadata, otherwise just write file data.
    fn fsync(&self, _req: RequestInfo, _path: &Path, _fh: u64, _datasync: bool) -> ResultEmpty {
        Err(libc::ENOSYS)
    }

    /// Open a directory.
    ///
    /// Analogous to the `opend` call.
    ///
    /// * `path`: path to the directory.
    /// * `flags`: file access flags. Will contain `O_DIRECTORY` at least.
    ///
    /// Return a tuple of (file handle, flags). The file handle will be passed to any subsequent
    /// calls that operate on the directory, and can be any value you choose, though it should
    /// allow your filesystem to identify the directory opened even without any path info.
    fn opendir(&self, _req: RequestInfo, path: &Path, flags: u32) -> ResultOpen {
        let path = path.to_str().unwrap();
        if let Some(_node) = self.vir_root.find(&path) {
            return Ok((6666, flags));
        }
        let request = tonic::Request::new(Opendir {
            path: path.to_string(),
            ..Opendir::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fopendir(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            return Err(resp.ret);
        }
        Ok((resp.fi.unwrap().fh, flags))
    }

    /// Get the entries of a directory.
    ///
    /// * `path`: path to the directory.
    /// * `fh`: file handle returned from the `opendir` call.
    ///
    /// Return all the entries of the directory.
    fn readdir(&self, _req: RequestInfo, path: &Path, fh: u64) -> ResultReaddir {
        fn filetype(kind: u32) -> FileType {
            match kind {
                0 => FileType::RegularFile,
                1 => FileType::Directory,
                2 => FileType::Symlink,
                3 => FileType::CharDevice,
                4 => FileType::BlockDevice,
                5 => FileType::NamedPipe,
                6 => FileType::Socket,
                _ => FileType::RegularFile,
            }
        }
        let path = path.to_str().unwrap();
        if let Some(node) = self.vir_root.find(&path) {
            let mut dirs = vec![];
            let resp = node.readdir();
            for name in resp {
                dirs.push(DirectoryEntry {
                    name: name.into(),
                    kind: FileType::Directory,
                });
            }
            return Ok(dirs);
        }
        let request = tonic::Request::new(Readdir {
            path: path.to_string(),
            fi: Some(FileInfo {
                fh,
                ..FileInfo::default()
            }),
            ..Readdir::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.freaddir(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            return Err(resp.ret)
        }
        let mut dirs = vec![];
        // let clrs: Vec<String> = resp.dirs.iter().map(|dir| dir.name.clone().into()).collect();
        // self.cache.clr(clrs);
        for dir in resp.dirs {
            dirs.push(DirectoryEntry {
                name: dir.name.into(),
                kind: filetype(dir.kind),
            });
        }
        Ok(dirs)
    }

    /// Close an open directory.
    ///
    /// This will be called exactly once for each `opendir` call.
    ///
    /// * `path`: path to the directory.
    /// * `fh`: file handle returned from the `opendir` call.
    /// * `flags`: the file access flags passed to the `opendir` call.
    fn releasedir(&self, _req: RequestInfo, path: &Path, fh: u64, _flags: u32) -> ResultEmpty {
        if fh == 6666 {
            return Ok(());
        }
        let request = tonic::Request::new(Releasedir {
            path: path.to_string_lossy().into_owned(),
            fi: Some(FileInfo {
                fh,
                ..FileInfo::default()
            }),
            ..Releasedir::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.freleasedir(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Write out any pending changes to a directory.
    ///
    /// Analogous to the `fsync` call.
    fn fsyncdir(&self, _req: RequestInfo, _path: &Path, _fh: u64, _datasync: bool) -> ResultEmpty {
        Err(libc::ENOSYS)
    }

    /// Get filesystem statistics.
    ///
    /// * `path`: path to some folder in the filesystem.
    ///
    /// See the `Statfs` struct for more details.
    fn statfs(&self, _req: RequestInfo, _path: &Path) -> ResultStatfs {
        Err(libc::ENOSYS)
    }

    /// Set a file extended attribute.
    ///
    /// * `path`: path to the file.
    /// * `name`: attribute name.
    /// * `value`: the data to set the value to.
    /// * `flags`: can be either `XATTR_CREATE` or `XATTR_REPLACE`.
    /// * `position`: offset into the attribute value to write data.
    fn setxattr(
        &self,
        _req: RequestInfo,
        _path: &Path,
        _name: &OsStr,
        _value: &[u8],
        _flags: u32,
        _position: u32,
    ) -> ResultEmpty {
        Err(libc::ENOSYS)
    }

    /// Get a file extended attribute.
    ///
    /// * `path`: path to the file
    /// * `name`: attribute name.
    /// * `size`: the maximum number of bytes to read.
    ///
    /// If `size` is 0, return `Xattr::Size(n)` where `n` is the size of the attribute data.
    /// Otherwise, return `Xattr::Data(data)` with the requested data.
    fn getxattr(&self, _req: RequestInfo, _path: &Path, _name: &OsStr, _size: u32) -> ResultXattr {
        Err(libc::ENOSYS)
    }

    /// List extended attributes for a file.
    ///
    /// * `path`: path to the file.
    /// * `size`: maximum number of bytes to return.
    ///
    /// If `size` is 0, return `Xattr::Size(n)` where `n` is the size required for the list of
    /// attribute names.
    /// Otherwise, return `Xattr::Data(data)` where `data` is all the null-terminated attribute
    /// names.
    fn listxattr(&self, _req: RequestInfo, _path: &Path, _size: u32) -> ResultXattr {
        Err(libc::ENOSYS)
    }

    /// Remove an extended attribute for a file.
    ///
    /// * `path`: path to the file.
    /// * `name`: name of the attribute to remove.
    fn removexattr(&self, _req: RequestInfo, _path: &Path, _name: &OsStr) -> ResultEmpty {
        Err(libc::ENOSYS)
    }

    /// Check for access to a file.
    ///
    /// * `path`: path to the file.
    /// * `mask`: mode bits to check for access to.
    ///
    /// Return `Ok(())` if all requested permissions are allowed, otherwise return `Err(EACCES)`
    /// or other error code as appropriate (e.g. `ENOENT` if the file doesn't exist).
    fn access(&self, _req: RequestInfo, path: &Path, mask: u32) -> ResultEmpty {
        fn mask2mode(mask: i32) -> u32 {
            let mut mode = 0;
            if mask & libc::F_OK != 0 {
                mode = 0;
            }
            if mask & libc::X_OK != 0 {
                mode |= 0o0111;
            }
            if mask & libc::W_OK != 0 {
                mode |= 0o222;
            }
            if mask & libc::R_OK != 0 {
                mode |= 0o444;
            }
            mode as u32
        }
        let request = tonic::Request::new(Access {
            path: path.to_string_lossy().into_owned(),
            mask: mask2mode(mask as i32),
            ..Access::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.faccess(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        ret2result!(resp.ret)
    }

    /// Create and open a new file.
    ///
    /// * `parent`: path to the directory to create the file in.
    /// * `name`: name of the file to be created.
    /// * `mode`: the mode to set on the new file.
    /// * `flags`: flags like would be passed to `open`.
    ///
    /// Return a `CreatedEntry` (which contains the new file's attributes as well as a file handle
    /// -- see documentation on `open` for more info on that).
    fn create(
        &self,
        req: RequestInfo,
        parent: &Path,
        name: &OsStr,
        mode: u32,
        flags: u32,
    ) -> ResultCreate {
        let path = format!("{}/{}", parent.to_str().unwrap(), name.to_str().unwrap());
        if let Some(_unused) = self.vir_root.find(&path) {
            return Err(libc::ENOMSG);
        }
        let request = tonic::Request::new(Create {
            path: path.clone(),
            mode,
            ..Create::default()
        });
        let (rpc, resp) = {
            let client = match self.rpcs.get() {
                Ok(c) => c,
                Err(_) => return Err(libc::EAGAIN),
            };
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.fcreate(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(libc::ENOMSG);
            }
        };
        log::trace!("get resp: {:?}", resp);
        if resp.ret != 0 {
            return Err(libc::ENOMSG);
        }
        self.cache.clr(vec![path.to_string()]);
        let fi = resp.fi.unwrap();
        match self.getattr(req, Path::new(&path), Some(fi.fh)) {
            Ok((ttl, attr)) => {
                Ok(CreatedEntry {
                    fh: fi.fh,
                    flags,
                    attr,
                    ttl,
                })
            },
            Err(e) => Err(e),
        }
    }
}

impl LwsVfsIns {
    pub async fn new(server: &String, cache_life: u32) -> Result<LwsVfsIns, Box<dyn Error>> {
        let addr = format!("http://{}", server);
        log::info!("lws server is {}", addr);
        let rpcs = RpcPool::new(10, addr).await?;
        let config = LwsVfsIns::fetch_config(&rpcs).await?;
        let mut vir_root = VirFs::new();
        for mount in config.get_mount() {
            vir_root.add(mount);
        }
        Ok(LwsVfsIns {
            config,
            rpcs,
            vir_root,
            cache: FileAttrCacheManager::new(cache_life),
        })
    }

    async fn fetch_config(rpcs: &RpcPool) -> Result<Config, Box<dyn Error>> {
        let (rpc, resp) = {
            let request = tonic::Request::new(GetConfig {
                ..GetConfig::default()
            });
            let client = rpcs.get().unwrap();
            let mut rpc = client.write().unwrap();
            let rpc = &mut (rpc.client);
            let resp = rpc.get_config(request).await;
            (client.clone(), resp)
        };
        rpcs.put(rpc);
        let resp = match resp {
            Ok(resp) => resp.into_inner(),
            Err(e) => {
                log::error!("get resp err: {:?}", e);
                return Err(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "get config err",
                )));
            }
        };
        let config = Config::new(&resp.config).expect("parse config err");
        log::info!("config: {:?}", config);
        Ok(config)
    }

    pub fn init(&self) -> Result<(), Box<dyn Error>> {
        let request = tonic::Request::new(HelloRequest {
            name: "Ekko lws hello".into(),
        });
        let (rpc, resp) = {
            let client = self.rpcs.get()?;
            let mut rpc = client.write().unwrap();
            let rt = rpc.rt.clone();
            let rpc = &mut (rpc.client);
            let resp = rt.block_on(rpc.say_hello(request));
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        log::trace!("resp={:?}", resp);
        Ok(())
    }

    pub async fn hello(&self) -> Result<(), Box<dyn Error>> {
        let request = tonic::Request::new(HelloRequest {
            name: "Ekko lws hello".into(),
        });
        let (rpc, resp) = {
            let client = self.rpcs.get()?;
            let mut rpc = client.write().unwrap();
            let resp = rpc.client.say_hello(request).await?;
            (client.clone(), resp)
        };
        self.rpcs.put(rpc);
        log::trace!("resp={:?}", resp);
        Ok(())
    }

    pub fn mount<F>(mount_point: &str, file_system: F) -> Result<(), Box<dyn Error>>
    where
        F: FilesystemMT + Sync + Send + 'static,
    {
        let user = match env::var("USER") {
            Ok(user) => user,
            Err(_) => {
                log::warn!("Can not get user name, use 'unknown'");
                "unknown".to_string()
            },
        };
        let name = format!("fsname=lws_vfs@{}", user);
        let fuse_args = [OsStr::new("-o"), OsStr::new(name.as_str())];
        fuse_mt::mount(
            fuse_mt::FuseMT::new(file_system, 10),
            &mount_point,
            &fuse_args[..],
        )?;
        Ok(())
    }
}
